/* globals user_pref */
user_pref("ldap_2.servers.extension.description", "extension");
user_pref("ldap_2.servers.extension.filename", "ldap1.mab");
user_pref(
  "ldap_2.servers.extension.uri",
  "ldap://test.invalid:389/o=invalid??sub"
);
